# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import logging
import scanpy as sc

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, Optional, Union


# -- Operational class: -------------------------------------------------------
class SCVerseNeighbors(ABCParse.ABCParse):
    def __init__(
        self,
        distances_key: str = "distances",
        connectivities_key: str = "connectivities",
        params_key: str = "neighbors",
        *args,
        **kwargs,
    ) -> None:
        """
        Args:
            distances_key (Optional[str]): Key accessor to cell neighbor
            distances in ``adata.obsp``. **Default**: "distances".
            connectivities_key (Optional[str]): Key accessor to cell neighbor
            connectivities in ``adata.obsp``. **Default**: "connectivities".
            params_key (str): **Default**: "neighbors".

        Returns:
            None
        """
        self.__parse__(locals())

        self._neighbor_computation_performed = False

    @property
    def _SCANPY_NEIGHBORS_KWARGS(self) -> Dict[str, Any]:
        return ABCParse.function_kwargs(sc.pp.neighbors, self._PARAMS)

    @property
    def _HAS_DISTANCES(self) -> bool:
        return self._distances_key in self._adata.obsp

    @property
    def _HAS_CONNECTIVITIES(self) -> bool:
        return self._connectivities_key in self._adata.obsp

    @property
    def _HAS_NN_PARAMS(self) -> bool:
        return self._params_key in self._adata.uns

    def _probable_fetch(self, key):
        return adata_query.fetch(self._adata, key)

    @property
    def distances(self):
        if self._HAS_DISTANCES:
            return self._probable_fetch(self._distances_key)

    @property
    def connectivities(self):
        if self._HAS_CONNECTIVITIES:
            return self._probable_fetch(self._connectivities_key)

    @property
    def params(self):
        if self._HAS_NN_PARAMS:
            return self._probable_fetch(self._params_key)

    @property
    def properties(self):
        return {
            attr[1:]: getattr(self, attr) for attr in self.__dir__() if "_HAS_" in attr
        }

    @property
    def neighbors_precomputed(self) -> bool:
        return all(self.properties.values())

    def _intake(self, adata: anndata.AnnData) -> None:
        """"""

        self._preexisting = {}

        for key in ["obsp", "uns"]:
            if not hasattr(adata, key):
                self._preexisting[key] = None
            else:
                self._preexisting[key] = list(getattr(adata, key).keys())

    def _message(self, adata: anndata.AnnData) -> None:
        for key, val in self._preexisting.items():
            added = [attr for attr in getattr(adata, key).keys() if not attr in val]
            for added_val in added:
                logger.info(f"Added: adata.{key}['{added_val}']")

    def forward(self, adata: anndata.AnnData, **kwargs) -> None:
        """"""
        self.__update__(kwargs)

        self._intake(adata)

        if self.neighbors_precomputed and self._force:
            logger.info("Force recomputing neighbors")

        if not self.neighbors_precomputed or self._force:
            kw = self._SCANPY_NEIGHBORS_KWARGS.copy()
            kw.update({"adata": adata})

            sc.pp.neighbors(**kw)

            self._neighbor_computation_performed = True
        self._message(adata)

    def __call__(
        self,
        adata: anndata.AnnData,
        n_neighbors: int = 15,
        n_pcs: Optional[int] = None,
        use_rep: Optional[str] = None,
        random_state: Optional[Union[int, None]] = 0,
        method: Optional[str] = 'umap',
        metric: Optional[str] = 'euclidean',
        distances_key: Optional[str] = "distances",
        connectivities_key: Optional[str] = "connectivities",
        params_key: Optional[str] = "neighbors",
        force: bool = False,
        silent: Optional[bool] = False,
        return_cls: Optional[bool] = False,
        *args,
        **kwargs,
    ) -> None:
        self.__update__(locals())

        self.forward(adata)


# -- API-facing function: -----------------------------------------------------
def scverse_neighbors(
    adata: anndata.AnnData,
    n_neighbors: int = 15,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    random_state: Optional[Union[int, None]] = 0,
    method: Optional = "umap",
    metric: Optional = "euclidean",
    distances_key: Optional[str] = "distances",
    connectivities_key: Optional[str] = "connectivities",
    params_key: Optional[str] = "neighbors",
    force: Optional[bool] = False,
    return_cls: Optional[bool] = False,
    *args,
    **kwargs,
):
    """Scanpy's ``sc.pp.neighbors`` with a few book-keeping functions added.

    Args:
        adata: anndata.AnnData

        n_neighbors (int): decsription. **Default** = 15

        distances_key (str): decsription. **Default** = "distances"

        connectivities_key (str): decsription. **Default** = "connectivities"

        params_key (str): decsription. **Default** = "neighbors"

        force (bool): decsription. **Default** = False

        return_cls (bool): Return the operator class for access to functional handles. **Default** = False

    Returns:
        None, updates ``adata``.
    """

    scv_neighbors = SCVerseNeighbors(
        distances_key=distances_key,
        connectivities_key=connectivities_key,
        params_key=params_key,
    )

    funcs = [sc.pp.neighbors, scv_neighbors.__call__]

    for func in funcs:
        kwargs.update(ABCParse.function_kwargs(func, locals()))

    scv_neighbors(**kwargs)

    if return_cls:
        return scv_neighbors




# from .. import _LOGGING

def _read_logs(logging = _LOGGING):
    with open(logging.log_config.log_fpath, "r") as f:
        _logs = f.readlines()
        f.close()
    return _logs

# logs = _read_logs()        
    





from ...core import utils

import pandas as pd
import numpy as np
import ABCParse


NoneType = type(None)

class Scatter3D(ABCParse.ABCParse):
    def __init__(self, z_key="psi", use_key="X_umap", z_adj=0.01, color=None):

        self.__parse__(locals(), public=[None])

    @property
    def X_umap(self):
        return self.adata.obsm[self._use_key]

    @property
    def Z_smooth(self) -> np.ndarray:
        return self.neighbor_model.predict(self.X_umap)

    def _compose_scatter_dict(self):

        SCATTER_DICT = {"UMAP-1": self.X_umap[:, 0], "UMAP-2": self.X_umap[:, 1]}
        SCATTER_DICT[self._z_key] = self.Z_smooth + self._z_adj

        if not isinstance(self._color, NoneType):
            SCATTER_DICT[self._color] = self.adata_scatter.obs[self._color]

        return SCATTER_DICT

    @property
    def scatter_frame(self) -> pd.DataFrame:
        return pd.DataFrame(self._compose_scatter_dict())

    def __call__(self, adata, neighbor_model):
        self.__update__(locals())
        return self.scatter_frame


from abc import abstractmethod
import glob, os
import pandas as pd

from .. import utils
import ABCParse

# -- Base Class: ------
class Logs(ABCParse.ABCParse):
    def __init__(self, path, stage, version=0):
        self.__parse__(locals(), public=[None])

    @property
    def _AVAILABLE_VERSIONS(self):
        versions = glob.glob(f"{self._path}/{self._stage}*/version_*")
        n_versions = len(versions)
        return [int(os.path.basename(v).split("_")[-1]) for v in versions]

    @property
    def _VERSION(self):
        """Sometimes the passed version isn't found..."""
        if self._version in self._AVAILABLE_VERSIONS:
            return self._version
        elif len(self._AVAILABLE_VERSIONS) > 0:
            return max(self._AVAILABLE_VERSIONS)
        else:
            return "VERSION NOT FOUND"

    @property
    def _BASE_PATH(self):
        return os.path.join(
            self._path, f"{self._stage}_logs", f"version_{self._VERSION}"
        )

    @property
    def _METRICS_PATH(self):
        _metrics_path = os.path.join(self._BASE_PATH, "metrics.csv")
        if os.path.exists(_metrics_path):
            return _metrics_path
#         else:
#             _metrics_path = os.path.join(self._BACKUP_BASE_PATH, "metrics.csv")
#             if os.path.exists(_metrics_path):
#                 return _metrics_path
#             raise FileNotFoundError(_metrics_path)

    @property
    def _HPARAMS_PATH(self):
        return os.path.join(self._BASE_PATH, "hparams.yaml")

    @property
    def _CKPTS_PATH(self):
        return os.path.join(self._BASE_PATH, "checkpoints/*")

    @property
    def METRICS(self):
        return pd.read_csv(self._METRICS_PATH)

    @property
    def HPARAMS(self):
        return sdq.tl.HyperParams(self._HPARAMS_PATH)

    @property
    def CKPTS(self):
        return glob.glob(self._CKPTS_PATH)

    @property
    def METRICS_BY_EPOCH(self):
        return self.METRICS.groupby("epoch")

    @property
    def _EPOCH_STEP_IDX(self):
        return self.METRICS[["epoch", "step"]]

    @property
    def _COLUMNS(self):
        return self.METRICS.columns.tolist()

    def _epoch_sum_mean_filtered_on(self, df, regex):
        _df = df[self._COLUMNS].copy()
        for re in regex:
            _df = _df.filter(regex=re)
        return _df.mean().sum()

    @abstractmethod
    def assemble(self):
        ...
        
    @property
    def _NON_ZERO_COLS(self):
        return self.df.columns[self.df.sum(0) > 0].tolist()

    @property
    def _NULL_COLS(self):
        return self.df.columns[self.df.sum(0) == 0].tolist()
    
    def __call__(self):
        
        self.df = self.assemble()
        return self.df[self._NON_ZERO_COLS]

class TrainLogs(Logs):
    def __init__(self, path, stage="train", version=0):
        super().__init__(path=path, stage=stage, version=version)

    # -- specialized:  everything above is consistent b/t pretrian/train.
    @property
    def TRAINING_SINKHORN(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["^(?!fate_weighted).*$", "sinkhorn", "training"],
            )
            .to_frame()
            .rename({0: "training_sinkhorn"}, axis=1)
        )

    @property
    def VALIDATION_SINKHORN(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["^(?!fate_weighted).*$", "sinkhorn", "validation"],
            )
            .to_frame()
            .rename({0: "validation_sinkhorn"}, axis=1)
        )

    @property
    def TRAINING_SINKHORN_FATE_WEIGHTED(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["fate_weighted", "sinkhorn", "training"],
            )
            .to_frame()
            .rename({0: "training_sinkhorn_fate_weighted"}, axis=1)
        )

    @property
    def VALIDATION_SINKHORN_FATE_WEIGHTED(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["fate_weighted", "sinkhorn", "validation"],
            )
            .to_frame()
            .rename({0: "validation_sinkhorn_fate_weighted"}, axis=1)
        )

    @property
    def TRAINING_FATE_ACCURACY(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["fate_acc_score", "training"],
            )
            .to_frame()
            .rename({0: "fate_accuracy_training"}, axis=1)
        )

    @property
    def VALIDATION_FATE_ACCURACY(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["fate_acc_score", "validation"],
            )
            .to_frame()
            .rename({0: "fate_accuracy_validation"}, axis=1)
        )

    @property
    def TRAINING_KL_DIVERGENCE(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["kl_div", "training"],
            )
            .to_frame()
            .rename({0: "kl_div_training"}, axis=1)
        )

    @property
    def VALIDATION_KL_DIVERGENCE(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on,
                ["kl_div", "validation"],
            )
            .to_frame()
            .rename({0: "kl_div_validation"}, axis=1)
        )

    def assemble(self):
        
        _df = pd.concat(
            [
                self.TRAINING_SINKHORN,
                self.VALIDATION_SINKHORN,
                self.TRAINING_SINKHORN_FATE_WEIGHTED,
                self.VALIDATION_SINKHORN_FATE_WEIGHTED,
                self.TRAINING_KL_DIVERGENCE,
                self.VALIDATION_KL_DIVERGENCE,
            ],
            axis=1,
        )
        _df["total_training"] = _df.filter(regex="training").sum(1)
        _df["total_validation"] = _df.filter(regex="validation").sum(1)

        # -- add accuracy scores after summing: -----
        _df["fate_accuracy_training"] = self.TRAINING_FATE_ACCURACY[
            "fate_accuracy_training"
        ].values
        _df["fate_accuracy_validation"] = self.VALIDATION_FATE_ACCURACY[
            "fate_accuracy_validation"
        ].values

        return _df
    



class PretrainLogs(Logs):
    def __init__(self, path, stage="pretrain", version=0):
        super().__init__(path=path, stage=stage, version=version)

    @property
    def PRETRAIN_RL_MSE(self):
        return (
            self.METRICS_BY_EPOCH.apply(
                self._epoch_sum_mean_filtered_on, ["pretrain_rl_mse"]
            )
            .to_frame()
            .rename({0: "pretrain_rl_mse"}, axis=1)
        )

    def assemble(self):
        """overkill but makes future extension easier"""
        return self.PRETRAIN_RL_MSE



# -- import packages: ---------------------------------------------------------
import ABCParse
import datetime
import lightning
import time

# -- set type hints: ----------------------------------------------------------
from typing import List, Optional

aliases = {"epoch_train_loss": "train loss", "epoch_validation_loss": "val loss"}


# -- callback cls: ------------------------------------------------------------
class BasicProgressBar(lightning.pytorch.callbacks.Callback):
    def __init__(self, total_epochs, metric_keys: Optional[List[str]] = []) -> None:
        """
        Args:
            total_epochs (int): total number of epochs
            metric_keys (list or None): which metrics to log (e.g., ["val_loss"])
        """
        self.total_epochs = total_epochs
        self.epoch_start_time = None
        self._metric_keys = ["epoch_train_loss", "epoch_validation_loss"]  # metric_keys

    @property
    def metric_keys(self):
        return ABCParse.as_list(self._metric_keys)

    @property
    def _now(self):
        return datetime.datetime.now().strftime("%H:%M:%S")

    def on_train_epoch_start(self, trainer, pl_module):
        self.epoch_start_time = time.time()

    def on_train_epoch_end(self, trainer, pl_module):
        current_epoch = trainer.current_epoch + 1
        elapsed = time.time() - self.epoch_start_time
        msg = f"[{self._now}] Epoch {current_epoch}/{self.total_epochs} | ({elapsed:.2f}s)"

        # Add metric logs if available
        metrics = trainer.callback_metrics
        metric_parts = []
        for key in self.metric_keys:
            val = metrics.get(key)
            if key in aliases:
                key = aliases[key]
            if val is not None:
                if isinstance(val, (float, int)):
                    metric_parts.append(f"{key}: {val:.2f}")
                else:
                    try:
                        val = float(val)
                        metric_parts.append(f"{key}: {val:.2f}")
                    except:
                        metric_parts.append(f"{key}: {val}")
        if metric_parts:
            msg += " | " + ", ".join(metric_parts)

        print(msg, flush=True)



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import lightning
import logging
import os
import pandas as pd
import pathlib
import torch
import warnings

# -- import local dependencies: -----------------------------------------------
from . import _mix_ins as mix_ins
from .. import __version__


# -- set type hints: ----------------------------------------------------------
from typing import Dict, List, Literal, Optional, Union


# -- setup logging: -----------------------------------------------------------
logger = logging.getLogger(__name__)

# -- remove specific unnecessary warnings from dependency: --------------------
warnings.filterwarnings(
    "ignore",
    ".*Consider increasing the value of the `num_workers` argument*",
)


# -- model class, faces API: --------------------------------------------------
class scDiffEq(
    mix_ins.ModelConfigMixIn,
    mix_ins.kNNMixIn,
    mix_ins.StateCharacterizationMixIn,
    mix_ins.UtilityIOMixIn,
    mix_ins.LoggingMixIn,
    mix_ins.PreTrainMixIn,
    mix_ins.TrainMixIn,
    ABCParse.ABCParse,
):
    """scDiffeq model class

    Parameters:
    ----------
    adata : Optional[anndata.AnnData], optional
        Annotated data matrix, by default None
    latent_dim : int, optional
        Number of latent dimensions, by default 50
    name : Optional[str], optional
        Model name, by default None
    use_key : str, optional
        Key to use for data, by default "X_pca"
    weight_key : str, optional
        Key to use for weights, by default "W"
    obs_keys : List[str], optional
        List of observation keys, by default []
    seed : int, optional
        Random seed, by default 0
    backend : str, optional
        Backend to use, by default "auto"
    gradient_clip_val : float, optional
        Gradient clipping value, by default 0.5
    velocity_ratio_params : Dict[str, Union[float, bool]], optional
        Parameters for velocity ratio, by default {"target": 2, "enforce": 100, "method": "square"}
    build_kNN : Optional[bool], optional
        Whether to build kNN, by default False
    kNN_key : Optional[str], optional
        Key to use for kNN, by default "X_pca"
    kNN_fit_subset : Optional[str], optional
        Subset to fit kNN, by default "train"
    pretrain_epochs : int, optional
        Number of pretrain epochs, by default 500
    pretrain_lr : float, optional
        Learning rate for pretraining, by default 1e-3
    pretrain_optimizer : torch.optim.Optimizer, optional
        Optimizer for pretraining, by default torch.optim.Adam
    pretrain_step_size : int, optional
        Step size for pretraining scheduler, by default 100
    pretrain_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
        Scheduler for pretraining, by default torch.optim.lr_scheduler.StepLR
    train_epochs : int, optional
        Number of training epochs, by default 1500
    train_lr : float, optional
        Learning rate for training, by default 1e-5
    train_optimizer : torch.optim.Optimizer, optional
        Optimizer for training, by default torch.optim.RMSprop
    train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
        Scheduler for training, by default torch.optim.lr_scheduler.StepLR
    train_step_size : int, optional
        Step size for training scheduler, by default 10
    train_val_split : List[float], optional
        Train-validation split, by default [0.9, 0.1]
    batch_size : int, optional
        Batch size, by default 2000
    train_key : str, optional
        Key for training data, by default "train"
    val_key : str, optional
        Key for validation data, by default "val"
    test_key : str, optional
        Key for test data, by default "test"
    predict_key : str, optional
        Key for prediction data, by default "predict"
    logger : Optional["lightning.logger"], optional
        Logger, by default None
    num_workers : int, optional
        Number of workers, by default os.cpu_count()
    silent : bool, optional
        Whether to silence output, by default True
    scale_input_counts : bool, optional
        Whether to scale input counts, by default False
    reduce_dimensions : bool, optional
        Whether to reduce dimensions, by default False
    fate_bias_csv_path : Optional[Union[pathlib.Path, str]], optional
        Path to fate bias CSV, by default None
    fate_bias_multiplier : float, optional
        Multiplier for fate bias, by default 1
    viz_frequency : int, optional
        Frequency of visualization, by default 1
    working_dir : Union[pathlib.Path, str], optional
        Working directory, by default os.getcwd()
    time_key : Optional[str], optional
        Key for time data, by default None
    t0_idx : Optional[pd.Index], optional
        Index for initial time, by default None
    t_min : float, optional
        Minimum time, by default 0
    t_max : float, optional
        Maximum time, by default 1
    dt : float, optional
        Time step, by default 0.1
    time_cluster_key : Optional[str], optional
        Key for time cluster, by default None
    t0_cluster : Optional[str], optional
        Initial time cluster, by default None
    shuffle_time_labels : bool, optional
        Whether to shuffle time labels, by default False
    mu_hidden : Union[List[int], int], optional
        Hidden layers for drift function, by default [400, 400]
    mu_activation : Union[str, List[str]], optional
        Activation function for drift, by default "LeakyReLU"
    mu_dropout : Union[float, List[float]], optional
        Dropout rate for drift, by default 0.1
    mu_bias : bool, optional
        Whether to use bias in drift, by default True
    mu_output_bias : bool, optional
        Whether to use output bias in drift, by default True
    mu_n_augment : int, optional
        Number of augmented dimensions for drift, by default 0
    sigma_hidden : Union[List[int], int], optional
        Hidden layers for diffusion function, by default [400, 400]
    sigma_activation : Union[str, List[str]], optional
        Activation function for diffusion, by default "LeakyReLU"
    sigma_dropout : Union[float, List[float]], optional
        Dropout rate for diffusion, by default 0.1
    sigma_bias : List[bool], optional
        Whether to use bias in diffusion, by default True
    sigma_output_bias : bool, optional
        Whether to use output bias in diffusion, by default True
    sigma_n_augment : int, optional
        Number of augmented dimensions for diffusion, by default 0
    adjoint : bool, optional
        Whether to use adjoint method, by default False
    sde_type : str, optional
        Type of SDE, by default "ito"
    noise_type : str, optional
        Type of noise, by default "general"
    brownian_dim : int, optional
        Dimension of Brownian motion, by default 1
    coef_drift : float, optional
        Coefficient for drift, by default 1.0
    coef_diffusion : float, optional
        Coefficient for diffusion, by default 1.0
    coef_prior_drift : float, optional
        Coefficient for prior drift, by default 1.0
    DiffEq_type : str, optional
        Type of differential equation, by default "SDE"
    potential_type : Union[None, str], optional
        Type of potential, by default None
    encoder_n_hidden : int, optional
        Number of hidden layers for encoder, by default 4
    encoder_power : float, optional
        Power for encoder, by default 2
    encoder_activation : Union[str, List[str]], optional
        Activation function for encoder, by default "LeakyReLU"
    encoder_dropout : Union[float, List[float]], optional
        Dropout rate for encoder, by default 0.2
    encoder_bias : bool, optional
        Whether to use bias in encoder, by default True
    encoder_output_bias : bool, optional
        Whether to use output bias in encoder, by default True
    decoder_n_hidden : int, optional
        Number of hidden layers for decoder, by default 4
    decoder_power : float, optional
        Power for decoder, by default 2
    decoder_activation : Union[str, List[str]], optional
        Activation function for decoder, by default "LeakyReLU"
    decoder_dropout : Union[float, List[float]], optional
        Dropout rate for decoder, by default 0.2
    decoder_bias : bool, optional
        Whether to use bias in decoder, by default True
    decoder_output_bias : bool, optional
        Whether to use output bias in decoder, by default True
    ckpt_path : Optional[Union[pathlib.Path, str]], optional
        Path to checkpoint, by default None
    version : str, optional
        Version of the model, by default __version__
    """

    def __init__(
        self,
        # -- data params: ------------------------------------------------------
        adata: Optional[anndata.AnnData] = None,
        latent_dim: int = 50,
        name: Optional[str] = None,
        use_key: str = "X_pca",
        weight_key: str = "W",
        obs_keys: List[str] = [],
        seed: int = 0,
        backend: str = "auto",
        gradient_clip_val: float = 0.5,
        # -- velocity ratio keys: ----------------------------------------------
        velocity_ratio_params: Dict[str, Union[float, bool]] = {
            "target": 2,
            "enforce": 100,  # zero to disable
            "method": "square",  # abs -> calls torch.abs or torch.square
        },
        # -- kNN keys: [optional]: ----------------------------------------------
        build_kNN: Optional[bool] = False,
        kNN_key: Optional[str] = "X_pca",
        kNN_fit_subset: Optional[str] = "train",
        # -- pretrain params: --------------------------------------------------
        pretrain_epochs: int = 500,
        pretrain_lr: float = 1e-3,
        pretrain_optimizer=torch.optim.Adam,
        pretrain_step_size: int = 100,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        # -- train params: -----------------------------------------------------
        train_epochs: int = 2500,
        train_lr: float = 1e-5,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        train_val_split: List[float] = [0.9, 0.1],
        batch_size: int = 2048,
        train_key: str = "train",
        val_key: str = "val",
        test_key: str = "test",
        predict_key: str = "predict",
        # -- general params: ---------------------------------------------------
        logger: Optional["lightning.logger"] = None,
        num_workers: int = 0,
        silent: bool = True,
        scale_input_counts: bool = False,
        reduce_dimensions: bool = False,
        fate_bias_csv_path: Optional[Union[pathlib.Path, str]] = None,
        fate_bias_multiplier: float = 1,
        viz_frequency: int = 1,
        working_dir: Union[pathlib.Path, str] = os.getcwd(),
        # -- time params: -------------------------------------------------------
        time_key: Optional[str] = None,
        t0_idx: Optional[pd.Index] = None,
        t_min: float = 0,
        t_max: float = 1,
        dt: float = 0.1,
        time_cluster_key: Optional[str] = None,
        t0_cluster: Optional[str] = None,
        shuffle_time_labels: bool = False,
        # -- DiffEq params: ---------------------------------------------------
        mu_hidden: Union[List[int], int] = [512, 512],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_hidden: Union[List[int], int] = [32, 32],
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_dropout: Union[float, List[float]] = 0.1,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        adjoint: bool = False,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        DiffEq_type: str = "SDE",
        potential_type: Literal["fixed", "prior"] = "fixed",
        # -- Encoder params: ---------------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- Decoder params: --------------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        ckpt_path: Optional[Union[pathlib.Path, str]] = None,
        monitor_hardware: bool = False,
        version: str = __version__,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the scDiffEq model.

        This class is responsible for bringing together three critical
        components: the lightning model, the lightning trainer, and the
        lightning data. All else is superfluous to model operation.

        Args:
            adata (Optional[anndata.AnnData], optional): Annotated data matrix. Default is None.
            latent_dim (int, optional): Number of latent dimensions. Default is 50.
            name (Optional[str], optional): Model name. Default is None.
            use_key (str, optional): Key to use for data. Default is "X_pca".
            weight_key (str, optional): Key to use for weights. Default is "W".
            obs_keys (List[str], optional): List of observation keys. Default is [].
            seed (int, optional): Random seed. Default is 0.
            backend (str, optional): Backend to use. Default is "auto".
            gradient_clip_val (float, optional): Gradient clipping value. Default is 0.5.
            velocity_ratio_params (Dict[str, Union[float, bool]], optional): Parameters for velocity ratio. Default is {"target": 2, "enforce": 100, "method": "square"}.
            build_kNN (Optional[bool], optional): Whether to build kNN. Default is False.
            kNN_key (Optional[str], optional): Key to use for kNN. Default is "X_pca".
            kNN_fit_subset (Optional[str], optional): Subset to fit kNN. Default is "train".
            pretrain_epochs (int, optional): Number of pretrain epochs. Default is 500.
            pretrain_lr (float, optional): Learning rate for pretraining. Default is 1e-3.
            pretrain_optimizer (torch.optim.Optimizer, optional): Optimizer for pretraining. Default is torch.optim.Adam.
            pretrain_step_size (int, optional): Step size for pretraining scheduler. Default is 100.
            pretrain_scheduler (torch.optim.lr_scheduler._LRScheduler, optional): Scheduler for pretraining. Default is torch.optim.lr_scheduler.StepLR.
            train_epochs (int, optional): Number of training epochs. Default is 1500.
            train_lr (float, optional): Learning rate for training. Default is 1e-5.
            train_optimizer (torch.optim.Optimizer, optional): Optimizer for training. Default is torch.optim.RMSprop.
            train_scheduler (torch.optim.lr_scheduler._LRScheduler, optional): Scheduler for training. Default is torch.optim.lr_scheduler.StepLR.
            train_step_size (int, optional): Step size for training scheduler. Default is 10.
            train_val_split (List[float], optional): Train-validation split. Default is [0.9, 0.1].
            batch_size (int, optional): Batch size. Default is 2000.
            train_key (str, optional): Key for training data. Default is "train".
            val_key (str, optional): Key for validation data. Default is "val".
            test_key (str, optional): Key for test data. Default is "test".
            predict_key (str, optional): Key for prediction data. Default is "predict".
            logger (Optional["lightning.logger"], optional): Logger. Default is None.
            num_workers (int, optional): Number of workers. Default is 0.
            silent (bool, optional): Whether to silence output. Default is True.
            scale_input_counts (bool, optional): Whether to scale input counts. Default is False.
            reduce_dimensions (bool, optional): Whether to reduce dimensions. Default is False.
            fate_bias_csv_path (Optional[Union[pathlib.Path, str]], optional): Path to fate bias CSV. Default is None.
            fate_bias_multiplier (float, optional): Multiplier for fate bias. Default is 1.
            viz_frequency (int, optional): Frequency of visualization. Default is 1.
            working_dir (Union[pathlib.Path, str], optional): Working directory. Default is os.getcwd().
            time_key (Optional[str], optional): Key for time data. Default is None.
            t0_idx (Optional[pd.Index], optional): Index for initial time. Default is None.
            t_min (float, optional): Minimum time. Default is 0.
            t_max (float, optional): Maximum time. Default is 1.
            dt (float, optional): Time step. Default is 0.1.
            time_cluster_key (Optional[str], optional): Key for time cluster. Default is None.
            t0_cluster (Optional[str], optional): Initial time cluster. Default is None.
            shuffle_time_labels (bool, optional): Whether to shuffle time labels. Default is False.
            mu_hidden (Union[List[int], int], optional): Hidden layers for drift function. Default is [400, 400].
            mu_activation (Union[str, List[str]], optional): Activation function for drift. Default is "LeakyReLU".
            mu_dropout (Union[float, List[float]], optional): Dropout rate for drift. Default is 0.1.
            mu_bias (bool, optional): Whether to use bias in drift. Default is True.
            mu_output_bias (bool, optional): Whether to use output bias in drift. Default is True.
            mu_n_augment (int, optional): Number of augmented dimensions for drift. Default is 0.
            sigma_hidden (Union[List[int], int], optional): Hidden layers for diffusion function. Default is [400, 400].
            sigma_activation (Union[str, List[str]], optional): Activation function for diffusion. Default is "LeakyReLU".
            sigma_dropout (Union[float, List[float]], optional): Dropout rate for diffusion. Default is 0.1.
            sigma_bias (List[bool], optional): Whether to use bias in diffusion. Default is True.
            sigma_output_bias (bool, optional): Whether to use output bias in diffusion. Default is True.
            sigma_n_augment (int, optional): Number of augmented dimensions for diffusion. Default is 0.
            adjoint (bool, optional): Whether to use adjoint method. Default is False.
            sde_type (str, optional): Type of SDE. Default is "ito".
            noise_type (str, optional): Type of noise. Default is "general".
            brownian_dim (int, optional): Dimension of Brownian motion. Default is 1.
            coef_drift (float, optional): Coefficient for drift. Default is 1.0.
            coef_diffusion (float, optional): Coefficient for diffusion. Default is 1.0.
            coef_prior_drift (float, optional): Coefficient for prior drift. Default is 1.0.
            DiffEq_type (str, optional): Type of differential equation. Default is "SDE".
            potential_type (Literal["fixed","prior"]): Type of potential. Default is "fixed".
            encoder_n_hidden (int, optional): Number of hidden layers for encoder. Default is 4.
            encoder_power (float, optional): Power for encoder. Default is 2.
            encoder_activation (Union[str, List[str]], optional): Activation function for encoder. Default is "LeakyReLU".
            encoder_dropout (Union[float, List[float]], optional): Dropout rate for encoder. Default is 0.2.
            encoder_bias (bool, optional): Whether to use bias in encoder. Default is True.
            encoder_output_bias (bool, optional): Whether to use output bias in encoder. Default is True.
            decoder_n_hidden (int, optional): Number of hidden layers for decoder. Default is 4.
            decoder_power (float, optional): Power for decoder. Default is 2.
            decoder_activation (Union[str, List[str]], optional): Activation function for decoder. Default is "LeakyReLU".
            decoder_dropout (Union[float, List[float]], optional): Dropout rate for decoder. Default is 0.2.
            decoder_bias (bool, optional): Whether to use bias in decoder. Default is True.
            decoder_output_bias (bool, optional): Whether to use output bias in decoder. Default is True.
            ckpt_path (Optional[Union[pathlib.Path, str]], optional): Path to checkpoint. Default is None.
            version (str, optional): Version of the model. Default is __version__.

        Returns: None
        """
        self.__config__(locals())

    def fit(
        self,
        train_epochs: int = 2500,
        pretrain_epochs: int = 0,
        train_lr: Optional[float] = None,
        pretrain_callbacks: List = [],
        train_callbacks: List = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor: Optional[Union[str, List[str]]] = None,
        accelerator: Optional[Union[str, List[str]]] = None,
        log_every_n_steps: int = 1,
        reload_dataloaders_every_n_epochs: int = 1,
        devices: int = 1,
        deterministic: bool = False,
        **kwargs: dict,
    ) -> None:
        """Fit the scDiffEq model to some data.

        Parameters
        ----------
        train_epochs : int, optional
            Number of training epochs, by default 200
        pretrain_epochs : int, optional
            Number of pretrain epochs, by default 500
        train_lr : float, optional
            Learning rate for training, by default None
        pretrain_callbacks : List, optional
            List of pretrain callbacks, by default []
        train_callbacks : List, optional
            List of train callbacks, by default []
        ckpt_frequency : int, optional
            Checkpoint frequency, by default 25
        save_last_ckpt : bool, optional
            Whether to save the last checkpoint, by default True
        keep_ckpts : int, optional
            Number of checkpoints to keep, by default -1
        monitor : optional
            Monitor for early stopping, by default None
        accelerator : optional
            Accelerator to use, by default None
        log_every_n_steps : int, optional
            Log every n steps, by default 1
        reload_dataloaders_every_n_epochs : int, optional
            Reload dataloaders every n epochs, by default 1
        devices : int, optional
            Number of devices to use, by default 1
        deterministic : bool, optional
            Whether to use deterministic algorithms, by default False

        Returns: None
        """
        self.train(**ABCParse.function_kwargs(self.train, locals()))

    def simulate(self) -> anndata.AnnData:
        """Simulate the scDiffEq model.

        Returns: anndata.AnnData
            Simulated data
        """
        ...

    def __repr__(self) -> str:
        """String representation of the scDiffEq model.
        Returns: str
        """
        if hasattr(self, "DiffEq"):
            return f"scDiffEq[{self.DiffEq.__repr__()}]"
        return "scDiffEq"



# -- import packages: ---------------------------------------------------------
import lightning
import logging
import pathlib
import pandas as pd

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- mix-in cls: --------------------------------------------------------------
class LoggingMixIn(object):
    def _configure_CSVLogger(self):
        """ """

        #         if hasattr(self, "_load_version"):
        #             version = self._load_version
        #         else:
        #             version = None

        return lightning.pytorch.loggers.CSVLogger(
            save_dir=self._working_dir,
            name=self._name,
            version=None,  # version,
            prefix="",
            flush_logs_every_n_steps=1,
        )

    @property
    def logger(self):
        if not hasattr(self, "_LOGGER"):
            self._csv_logger = self._configure_CSVLogger()
            if hasattr(self, "_logger") and (not self._logger is None):
                self._LOGGER = [self._csv_logger, self._logger]
            else:
                self._LOGGER = [self._csv_logger]

        return self._LOGGER

    @property
    def version(self) -> int:
        return self.logger[0].version

    @property
    def _metrics_path(self):
        return pathlib.Path(self.logger[0].log_dir).joinpath("metrics.csv")

    @property
    def metrics(self):
        return pd.read_csv(self._metrics_path)




from ._scdiffeq import scDiffEq

from . import lightning_models
from . import utils
from . import configs
from . import callbacks

__all__ = [
    "scDiffEq", "lightning_models", "callbacks", "configs", "utils", 
]



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins

# -- set type hints: ----------------------------------------------------------
from typing import List, Optional, Union

# -- lightning model: ---------------------------------------------------------
class LightningODE(
    mix_ins.BaseForwardMixIn,
    base.BaseLightningDiffEq,
):
    """LightningODE"""

    def __init__(
        self,
        # -- ode params: ------------------------------------------------------
        latent_dim: int = 50,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        backend: str = "auto",
        # -- general params: --------------------------------------------------
        train_lr: float = 1e-4,
        train_optimizer: torch.optim.Optimizer = torch.optim.RMSprop,
        train_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint: bool=False,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE

        Extended description.

        Parameters
        ----------
        latent_dim : int, optional
            Description. Default is 50.
        name : str, optional
            Description. Default is None.
        mu_hidden : Union[List[int], int], optional
            Description. Default is [2000, 2000].
        mu_activation : Union[str, List[str]], optional
            Description. Default is 'LeakyReLU'.
        mu_dropout : Union[float, List[float]], optional
            Description. Default is 0.2.
        mu_bias : bool, optional
            Description. Default is True.
        mu_output_bias : bool, optional
            Description. Default is True.
        mu_n_augment : int, optional
            Description. Default is 0.
        sde_type : str, optional
            Description. Default is 'ito'.
        noise_type : str, optional
            Description. Default is 'general'.
        backend : str, optional
            Description. Default is "auto".
        train_lr : float, optional
            Description. Default is 1e-4.
        train_optimizer : torch.optim.Optimizer, optional
            Description. Default is torch.optim.RMSprop.
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Description. Default is torch.optim.lr_scheduler.StepLR.
        train_step_size : int, optional
            Description. Default is 10.
        dt : float, optional
            Description. Default is 0.1.
        adjoint : bool, optional
            Description. Default is False.
        version : str, optional
            Description. Default is __version__.
        loading_existing : bool, optional
            Whether to load an existing model. Default is False.
        Returns
        -------
        None
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()

        # -- torch modules: ---------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.NeuralODE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

    def __repr__(self) -> str:
        return "LightningODE"



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import cellplots
import logging
import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import os
import pathlib
import scdiffeq_plots as sdq_pl

# -- import local dependencies: -----------------------------------------------
from ..core import utils

# -- set type hints: ----------------------------------------------------------
from typing import Dict, List, Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- cls: ---------------------------------------------------------------------
class FillBetweenPlot:
    """KNOWN BUG: CELLPLOTS REQUIRED / NOT YET RELEASED"""

    def __init__(self, *args, **kwargs): ...

    @property
    def _MEAN(self):
        return self._df["mean"]

    @property
    def _STD(self):
        return self._df["std"]

    @property
    def X(self):
        return self._MEAN.index

    @property
    def Y(self):
        return self._MEAN

    @property
    def YMIN(self):
        return self._MEAN - self._STD

    @property
    def YMAX(self):
        return self._MEAN + self._STD

    def _configure_plot(self, ax):
        if ax is None:
            # KNOWN BUG
            self.fig, axes = cellplots.plot(1, 1, delete=[["top", "right"]])
            ax = axes[0]
        return ax

    def _plot(self, label=None, *args, **kwargs):
        self.ax.fill_between(x=self.X, y1=self.YMIN, y2=self.YMAX, alpha=0.2, zorder=0)
        self.ax.plot(self.X, self.Y, zorder=1, label=label)

    def __call__(
        self,
        df,
        ax: Optional[matplotlib.axes.Axes] = None,
        label: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        self._df = df
        self.ax = self._configure_plot(ax)
        self._plot(label, **kwargs)


class SmoothedGEXPlot(ABCParse.ABCParse):
    def __init__(
        self,
        smoothed_gex_key: str = "t_smoothed_gex",
        nplots=1,
        ncols=1,
        xy_spines=True,
        height=0.5,
        width=0.5,
        ticklabel_size=6,
        x_label: Optional[str] = None,
        y_label: str = "Z-score",
        label_fontsize: float = 8,
        legend_loc: tuple = (1, 0.4),
        legend_fontsize: float = 8,
        *args,
        **kwargs,
    ) -> None:
        self.__parse__(locals(), public=[None])

    @property
    def _PLOT_KWARGS(self):
        return utils.function_kwargs(func=sdq_pl.plot, kwargs=self._PARAMS)

    def _build_plot(self):
        self.fig, self.axes = sdq_pl.plot(**self._PLOT_KWARGS)

    def _formatting(self, ax):
        ax.tick_params(axis="both", which="both", labelsize=self._ticklabel_size)
        ax.set_xlabel(self._X_LABEL, fontsize=self._label_fontsize)
        ax.set_ylabel(self._y_label, fontsize=self._label_fontsize)
        ax.legend(
            facecolor="None",
            edgecolor="None",
            loc=self._legend_loc,
            fontsize=self._legend_fontsize,
        )

    @property
    def _SMOOTHED_GEX(self) -> Dict:
        return self.adata_sim.uns[self._smoothed_gex_key]

    def forward(self, gene, gene_df):

        fill_between_plot = FillBetweenPlot()
        fill_between_plot(gene_df, ax=self.axes[0], label=gene)

    @property
    def _X_LABEL(self):
        if self._x_label is None:
            return self._smoothed_gex_key.split("_")[0]
        else:
            return self._x_label

    @property
    def _SAVE_PATH(self):
        if self._save_path is None:
            return pathlib.Path(
                os.path.join(
                    os.getcwd(),
                    "scDiffEq_figures",
                    ".".join(
                        [
                            "smoothed_gex",
                            "_".join(list(self._SMOOTHED_GEX.keys())),
                            self._X_LABEL,
                            "svg",
                        ],
                    ),
                )
            )
        return self._save_path

    def __call__(
        self,
        adata_sim: anndata.AnnData,
        save: bool = False,
        save_path: Optional[Union[pathlib.Path, str]] = None,
        *args,
        **kwargs,
    ):

        self.__update__(locals(), private=["save"])
        self._save_path = save_path
        self._build_plot()

        for gene, gene_df in self._SMOOTHED_GEX.items():
            self.forward(gene, gene_df)

        self._formatting(ax=self.axes[0])

        if self._save:
            if not self._SAVE_PATH.parent.exists():
                self._SAVE_PATH.parent.mkdir()
                logger.info(f"Directory created: {self._SAVE_PATH.parent}")
            plt.savefig(self._SAVE_PATH)
            logger.info(f"Saved to: {self._SAVE_PATH}")


def plot_smoothed_expression(
    adata_sim: anndata.AnnData,
    smoothed_gex_key: str = "t_smoothed_gex",
    save: bool = False,
    save_path: Optional[Union[pathlib.Path, str]] = None,
    nplots=1,
    ncols=1,
    xy_spines=True,
    height=0.5,
    width=0.5,
    ticklabel_size=6,
    x_label: Optional[str] = None,
    y_label: str = "Z-score",
    label_fontsize: float = 8,
    legend_loc: tuple = (1, 0.4),
    legend_fontsize: float = 8,
    *args,
    **kwargs,
):
    """
    Parameters
    ----------
    adata_sim: anndata.AnnData

    smoothed_gex_key: str = "t_smoothed_gex",

    nplots=1,

    ncols=1,

    xy_spines=True,

    height=0.5,

    width=0.5,

    ticklabel_size=6,

    x_label: Optional[str] = None,

    y_label: str = "Z-score",

    label_fontsize: float = 8,

    legend_loc: tuple = (1, 0.4),

    legend_fontsize: float = 8,

    Returns
    -------

    Notes
    -----

    """
    pl_smooth_gex = SmoothedGEXPlot(**locals())
    pl_smooth_gex(adata_sim, save=save, save_path=save_path)



# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib
import pandas as pd


# -- import local dependencies: -----------------------------------------------
from ._hparams import HParams
from ._checkpoint import Checkpoint
from ._grouped_metrics import GroupedMetrics


# -- set typing: --------------------------------------------------------------
from typing import Dict, List, Union


# -- operational class: -------------------------------------------------------
class Version(ABCParse.ABCParse):
    """scDiffEq Version object container

    Attributes
    ----------
    _path : Union[pathlib.Path, str]
        Path to the version within an scDiffEq project.
    _groupby : str
        Grouping method for metrics, default is "epoch".
    """

    def __init__(
        self,
        path: Union[pathlib.Path, str] = None,
        groupby: str = "epoch",
        *args,
        **kwargs
    ):
        """Instantiate Version by providing a path

        Parameters
        ----------
        path : Union[pathlib.Path, str], optional
            Path to the version within an scDiffEq project, by default None.
        groupby : str, optional
            Grouping method for metrics, by default "epoch"

        Returns
        -------
        None
        """
        self.__parse__(locals())

    @property
    def _PATH(self) -> pathlib.Path:
        """Check and format the provided path

        Returns
        -------
        pathlib.Path
            Formatted path

        Raises
        ------
        TypeError
            If the path is not of type pathlib.Path or str
        """
        if isinstance(self._path, pathlib.Path):
            return self._path
        if isinstance(self._path, str):
            self._path = pathlib.Path(self._path)
        return self._path

    @property
    def _NAME(self) -> str:
        """Version name from provided path

        Returns
        -------
        str
            Name of the version
        """
        return self._PATH.name

    @property
    def _CONTENTS(self) -> List[pathlib.Path]:
        """Return one-level glob of the provided path

        Returns
        -------
        List[pathlib.Path]
            List of contents in the provided path
        """
        return list(self._PATH.glob("*"))

    @property
    def hparams(self):
        """Check if the .yaml exists and instantiate the HParams class each time

        Returns
        -------
        HParams
            Instance of HParams class if hparams.yaml exists
        """
        hparams_path = self._PATH.joinpath("hparams.yaml")
        if hparams_path.exists():
            return HParams(hparams_path)

    @property
    def metrics_df(self) -> pd.DataFrame:
        """Check if metrics.csv path exists and read it

        Returns
        -------
        pd.DataFrame
            DataFrame containing the metrics if metrics.csv exists
        """
        metrics_path = self._PATH.joinpath("metrics.csv")
        if metrics_path.exists():
            return pd.read_csv(metrics_path)

    @property
    def per_epoch_metrics(self):
        """Group metrics by the specified groupby attribute

        Returns
        -------
        GroupedMetrics
            Instance of GroupedMetrics class
        """
        self._GROUPED_METRICS = GroupedMetrics(groupby=self._groupby)
        return self._GROUPED_METRICS(self.metrics_df)

    @property
    def _CKPT_PATHS(self) -> List[pathlib.Path]:
        """Formatted checkpoint paths

        Returns
        -------
        List[pathlib.Path]
            List of checkpoint paths
        """
        _ckpt_paths = list(self._PATH.joinpath("checkpoints").glob("*"))
        return [pathlib.Path(path) for path in _ckpt_paths]

    @property
    def _SORTED_CKPT_KEYS(self) -> List:
        """Sorting for organization's sake

        Returns
        -------
        List
            Sorted list of checkpoint keys
        """
        epochs = list(self.ckpts.keys())
        _epochs = sorted([epoch for epoch in epochs if epoch != "last"])
        if "last" in epochs:
            _epochs.append("last")
        return _epochs

    @property
    def ckpts(self) -> Dict:
        """Format and update available checkpoints for the version

        Returns
        -------
        Dict
            Dictionary of checkpoints
        """
        if not hasattr(self, "_CHECKPOINTS"):
            self._CHECKPOINTS = {}
            for ckpt_path in self._CKPT_PATHS:
                ckpt = Checkpoint(ckpt_path)
                self._CHECKPOINTS[ckpt.epoch] = ckpt
        return self._CHECKPOINTS

    def __repr__(self) -> str:
        """Return the name of the object

        Returns
        -------
        str
            Name of the object
        """
        return self._NAME



# -- import packages: ----------------------------------------------------------
import gc
import lightning
import os
import torch


# -- operational cls: ----------------------------------------------------------
class MemoryMonitor(lightning.pytorch.callbacks.Callback):

    def __init__(self, log_gpu=True):

        import psutil

        super().__init__()

        self.log_gpu = log_gpu
        self.process = psutil.Process(os.getpid())

    def _log_cpu_ram(self, trainer):

        import wandb

        mem = self.process.memory_info().rss / 1024**2  # in MB
        print(f"[MemoryLogger] Epoch {trainer.current_epoch}: CPU RAM: {mem:.2f} MB")
        wandb.log({"Memory/CPU_RAM_MB": mem})

    def _log_gpu_ram(self, trainer, lit_module):
        if self.log_gpu and lit_module.device.type == "cuda":
            gpu_mem_allocated = lit_module.device
            mem_allocated = torch.cuda.memory_allocated(lit_module.device) / 1024**2
            mem_reserved = torch.cuda.memory_reserved(lit_module.device) / 1024**2
            print(
                f"[MemoryLogger] Epoch {trainer.current_epoch}: GPU Allocated: {mem_allocated:.2f} MB | Reserved: {mem_reserved:.2f} MB"
            )
            wandb.log(
                {
                    "Memory/GPU_Allocated_MB": mem_allocated,
                    "Memory/GPU_Reserved_MB": mem_reserved,
                }
            )

    def _report_object_memories(self, trainer):
        objs = gc.get_objects()
        print(
            f"[MemoryLogger] Epoch {trainer.current_epoch}: Number of tracked objects: {len(objs)}"
        )

    def on_train_epoch_end(self, trainer, lit_module):

        self._log_cpu_ram(trainer=trainer)
        self._log_gpu_ram(trainer=trainer, lit_module=lit_module)
        self._report_object_memories(trainer=trainer)




import numpy as np


def get_iterative_indices(
    indices,
    index,
    n_recursive_neighbors: int = 2,
    max_neighs=None,
):
    """Iterative index searching"""

    def iterate_indices(indices, index, n_recursive_neighbors):
        if n_recursive_neighbors > 1:
            index = iterate_indices(indices, index, n_recursive_neighbors - 1)
        ix = np.append(index, indices[index])  # direct and indirect neighbors
        if np.isnan(ix).any():
            ix = ix[~np.isnan(ix)]
        return ix.astype(int)

    indices = np.unique(iterate_indices(indices, index, n_recursive_neighbors))
    if max_neighs is not None and len(indices) > max_neighs:
        indices = np.random.choice(indices, max_neighs, replace=False)
    return indices



# -- import packages: ----------------------------------------------------------
import ABCParse
import lightning
import logging
import os
import torch

# -- import local dependencies: ------------------------------------------------
from ._lightning_callbacks_configuration import LightningCallbacksConfiguration
from ._progress_bar_config import ProgressBarConfig

# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- Main class: --------------------------------------------------------------
class LightningTrainerConfiguration(ABCParse.ABCParse):
    def __init__(self, save_dir: str = "scDiffEq_Model"):
        super().__init__()

        self.__parse__(locals())
        if not os.path.exists(self._save_dir):
            os.mkdir(self._save_dir)

    # -- kwargs: ---------------------------------------------------------------
    #     @property
    #     def _CSVLogger_kwargs(self):
    #         return ABCParse.function_kwargs(
    #             func=loggers.CSVLogger, kwargs=self._PARAMS, ignore=['version'],
    #         )

    @property
    def _Trainer_kwargs(self):
        ignore = ["accelerator", "callbacks", "version", "logger"]
        tk = ABCParse.function_kwargs(
            func=lightning.pytorch.Trainer,
            kwargs=self._PARAMS,
        )
        for val in ignore:
            if val in tk.keys():
                tk.pop(val)
        return tk

    @property
    def Callbacks(self):
        callback_config = LightningCallbacksConfiguration()

        callbacks = callback_config(
            version=self._version,
            monitor_hardware=self._monitor_hardware,
            stage=self._stage,
            viz_frequency=self._viz_frequency,
            model_name=self._model_name,
            working_dir=self._working_dir,
            train_version=self._train_version,
            pretrain_version=self._pretrain_version,
            callbacks=self._callbacks,
            ckpt_frequency=self._ckpt_frequency,
            keep_ckpts=self._keep_ckpts,
            monitor=self._monitor,
            retain_test_gradients=self._retain_test_gradients,
            save_last=self._save_last_ckpt,
            # swa_lrs=1e-5,
        )
        callbacks.extend(self._progress_bar_config.pbar)
        return callbacks

    #         return callback_config(
    #             callbacks=self._callbacks,
    #             ckpt_frequency=self.ckpt_frequency,
    #             keep_ckpts=self.keep_ckpts,
    #             retain_test_gradients=self.retain_test_gradients,
    #             monitor = self.monitor,
    # #             swa_lrs = self.swa_lrs,
    #             save_last = self.save_last_ckpt,
    #             version = self.version,
    #         )

    @property
    def accelerator(self):
        if not self._accelerator is None:
            return self._accelerator
        if torch.cuda.is_available():
            return "gpu"
        if torch.backends.mps.is_available():  # experimental
            return "mps"
        return "cpu"

    # -- trainers: -------------------------------------------------------------
    @property
    def Trainer(self):
        """
        Main Lightning Trainer used for fitting / testing.
        If pre-train routine was used, Trainer loads from ckpt path.
        """

        return lightning.pytorch.Trainer(
            accelerator=self.accelerator,
            logger=self._logger,  # loggers.CSVLogger(**self._CSVLogger_kwargs),
            callbacks=self.Callbacks,
            enable_progress_bar=self._progress_bar_config.enable_progress_bar,
            **self._Trainer_kwargs,
        )

    @property
    def GradientsRetainedTestTrainer(self):
        """
        Quasi test trainer - serves as a workaround for evaluating test data
        while retaining gradients.
        """

        self._Trainer_kwargs["max_epochs"] = 0
        self._Trainer_kwargs["callbacks"] = self.Callbacks

        return lightning.pytorch.Trainer(
            accelerator=self.accelerator,
            logger=self._logger,  # loggers.CSVLogger(**self._CSVLogger_kwargs),
            num_sanity_val_steps=-1,
            enable_progress_bar=self._progress_bar_config.enable_progress_bar,
            **self._Trainer_kwargs,
        )

    def __call__(
        self,
        logger,
        lr: float = None,
        model_name="scDiffEq_model",
        gradient_clip_val: float = 0.5,
        monitor_hardware: bool = False,
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
        viz_frequency=1,
        stage=None,
        max_epochs=500,
        monitor=None,
        accelerator=None,
        devices=1,
        prefix: str = "",
        log_every_n_steps=1,
        flush_logs_every_n_steps: int = 1,
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        version: Optional[Union[int, str]] = None,
        callbacks: list = [],
        potential_model: bool = False,
        check_val_every_n_epoch=1,
        limit_val_batches=None,
        num_sanity_val_steps=None,
        val_check_interval=None,
        #         swa_lrs: float = None,
        reload_dataloaders_every_n_epochs=1,
        **kwargs,
    ):
        """

        Args:
        stage (str): stage of the training.
        max_epochs (int): maximum number of epochs to train.
        callbacks (list): list of callbacks to use.
        potential_model (bool): whether the model is a potential model.
        check_val_every_n_epoch (int): number of epochs to check validation.
        limit_val_batches (int): number of validation batches to use.
        num_sanity_val_steps (int): number of sanity validation steps.
        val_check_interval (int): number of epochs to check validation.
        reload_dataloaders_every_n_epochs (int): number of epochs to reload dataloaders.
        **kwargs: additional keyword arguments to pass to the Trainer.

        Returns: TrainerGenerator
        """

        self._gradient_clip_val = gradient_clip_val
        self._logger = logger

        self._retain_test_gradients = False
        if stage is None:
            stage = ""
        self._stage = stage
        self._progress_bar_config = ProgressBarConfig(total_epochs=max_epochs)

        if torch.cuda.device_count() > 0:
            devices = torch.cuda.device_count()

        self.__parse__(locals())

        if (potential_model) and (stage in ["test", "predict"]):
            self._retain_test_gradients = True
            return self.GradientsRetainedTestTrainer

        return self.Trainer



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins
from .. import utils

# -- set type hints: ----------------------------------------------------------
from typing import Literal, Optional, Union, List

# -- lightning model class: ---------------------------------------------------
class LightningODE_PriorPotential(
    mix_ins.PotentialMixIn,
    mix_ins.DriftPriorMixIn,
    base.BaseLightningDiffEq,
):
    """LightningODE-PriorPotential"""
    def __init__(
        self,
        # -- general params: --------------------------------------------------
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr: float = 1e-5,
        train_optimizer: torch.optim.Optimizer = torch.optim.RMSprop,
        train_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        adjoint: bool = False,
        # -- ODE params: ------------------------------------------------------
        coef_diff: float = 0,
        dt: float = 0.1,
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.2,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sde_type: str = "ito",
        noise_type: str = "general",
        backend: str = "auto",
        brownian_dim: int = 1,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningODE_PriorPotential

        Parameters:
        -----------
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50
        name : str, optional
            Name of the model, by default None
        train_lr : float, optional
            Learning rate for training, by default 1e-5
        train_optimizer : torch.optim.Optimizer, optional
            Optimizer for training, by default torch.optim.RMSprop
        train_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
            Learning rate scheduler for training, by default torch.optim.lr_scheduler.StepLR
        train_step_size : int, optional
            Step size for the learning rate scheduler, by default 10
        adjoint : bool, optional
            Whether to use the adjoint method for the ODE solver, by default False
        coef_diff : float, optional
            Coefficient of diffusion, by default 0
        dt : float, optional
            Time step for the ODE solver, by default 0.1
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the neural network, by default [400, 400]
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the neural network, by default 'LeakyReLU'
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the neural network, by default 0.2
        mu_bias : bool, optional
            Whether to use bias in the neural network, by default True
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the neural network, by default True
        mu_n_augment : int, optional
            Number of augmentations for the neural network, by default 0
        sde_type : str, optional
            Type of stochastic differential equation, by default 'ito'
        noise_type : str, optional
            Type of noise, by default 'general'
        backend : str, optional
            Backend for the ODE solver, by default "auto"
        brownian_dim : int, optional
            Dimensionality of the Brownian motion, by default 1
        loading_existing : bool, optional
            Whether to load an existing model, by default False

        Returns:
        --------
        None

        Notes:
        ------
        This class implements a prior potential ODE using PyTorch Lightning.

        Examples:
        ---------
        >>> model = LightningODE_PriorPotential(latent_dim=20, dt=0.05)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters()
        self.func = LatentPotentialODE(
            state_size=latent_dim,
            **utils.extract_func_kwargs(func=neural_diffeqs.LatentPotentialODE, kwargs=locals()),
        )
        self._configure_lightning_model(kwargs=locals())

    def forward(self, X0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        Z_hat, KL_div = self.integrate(
            Z0=X0, t=t, dt=self.hparams["dt"], logqp=True, **kwargs
        )
        return Z_hat, KL_div

    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):

        sinkhorn_loss = self.log_sinkhorn_divergence(sinkhorn_loss).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )

    def __repr__(self) -> Literal['LightningODE-PriorPotential']:
        return "LightningODE-PriorPotential"



# -- import packages: --------------------------------------------------------
import anndata
import lightning
import logging

# -- import local dependencies: -----------------------------------------------
from .. import configs, utils

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -- set type hints: ----------------------------------------------------------
from typing import Dict, Optional

# -- configure logging: -------------------------------------------------------
logger = logging.getLogger(__name__)


# -- mix-in cls: --------------------------------------------------------------
class ModelConfigMixIn(object):
    """configure_model and configure_data can/should be accessed from the public handle"""

    def _configure_trainer_generator(self) -> None:
        """ """

        self.TrainerGenerator = configs.LightningTrainerConfiguration(
            save_dir=self._name,
        )
        logger.debug(f"TrainerGenerator configured: {self._name}")

    def configure_model(
        self,
        DiffEq: Optional[lightning.LightningModule] = None,
        configure_trainer: bool = True,
        loading_existing: bool = False,
    ) -> None:
        if DiffEq is None:
            self._LitModelConfig = configs.LightningModelConfiguration(
                data_dim=self._data_dim,
                latent_dim=self._latent_dim,
                DiffEq_type=self._DiffEq_type,
                potential_type=self._potential_type,
                fate_bias_csv_path=self._fate_bias_csv_path,
                velocity_ratio_params=self._velocity_ratio_params,
            )

            if hasattr(self, "reducer"):
                self._PARAMS["PCA"] = self.reducer.PCA
            DiffEq = self._LitModelConfig(
                self._PARAMS, self._ckpt_path, loading_existing=loading_existing
            )

        self.DiffEq = DiffEq
        self._name = self.DiffEq.hparams.name

        logger.debug(f"Using the specified parameters, {self.DiffEq} has been called.")
        self._component_loader = utils.FlexibleComponentLoader(self)

        lightning.seed_everything(self._seed)

        # -- Step 5: configure bridge to lightning logger ----------------------
        # was its own step before: now in-line here, since it
        # doesn't make sense to separate it, functionally
        self._LOGGING = utils.LoggerBridge(self.DiffEq)

        if configure_trainer:
            self._configure_trainer_generator()

    def configure_data(self, adata: anndata.AnnData) -> None:
        """ """
        self.adata = adata.copy()

        self._DATA_CONFIG = configs.DataConfiguration()
        self._DATA_CONFIG(scDiffEq=self)
        logger.info(f"Input data configured.")

    def __config__(self, kwargs: Dict):
        """ """
        # -- Step 1: parse kwargs, set up info msg -----------------------------
        self.__parse__(kwargs, public=[None], ignore=["adata"])

        # -- Step 2: configure data ----------------------------------------------
        if not kwargs["adata"] is None:
            # if adata is given, triggers 2 through 4
            self.configure_data(adata=kwargs["adata"])

            # -- Step 3: configure kNN --------------------------------------------
            if self._PARAMS["build_kNN"]:
                self._PARAMS["kNN"] = self.kNN

            # -- Step 4: configure model -------------------------------------------
            self.configure_model(DiffEq=None, configure_trainer=True)

            # -- Step 5: extras (was step 6): ---------------------------------------


#             if kwargs["reduce_dimensions"]:
#                 self._configure_dimension_reduction()



# -- import packages: ----------------------------------------------------------
import anndata
import annoyance
import logging
import numpy as np
import pandas as pd
import sklearn.decomposition
import torch

# -- import local dependencies: ------------------------------------------------
from ._anndata_inspector import AnnDataInspector

# -- set type hints: -----------------------------------------------------------
from typing import Optional

# -- configure logging: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- operational cls: ----------------------------------------------------------
class FastGraph:
    """ """

    def __init__(
        self,
        adata: anndata.AnnData,
        use_key: str,
        annot_key: str = "Cell type annotation",
    ) -> None:
        """ """

        self.adata = adata
        self._inspector = AnnDataInspector(adata)
        self.annot_key = annot_key
        self.use_key = use_key
        if use_key in self._inspector.layers:
            adata.obsm[use_key] = adata.layers[use_key]

        self.Graph = annoyance.kNN(adata, use_key=self.use_key)
        self.Graph.build()

    def _fast_count(self, X_nn):
        return (
            self.adata[X_nn.flatten().astype(str)]
            .obs[self.annot_key]
            .values.reshape(-1, self.Graph._n_neighbors)
        )

    def _query(self, X_fin):
        return self._fast_count(self.Graph.query(X_fin))

    def _fate_df(self, x_lab) -> pd.DataFrame:
        return pd.DataFrame([x_lab[0].value_counts() for i in range(len(x_lab))])

    def _DETACH(self, X_hat: torch.Tensor) -> np.ndarray:
        return X_hat.detach().cpu().numpy()

    def _DETACH_FINAL(self, X_hat: torch.Tensor) -> np.ndarray:
        return X_hat[-1].detach().cpu().numpy()

    def _TRANSFORM(self, dimension_reduction_model, X_fin):
        return dimension_reduction_model.transform(X_fin)

    def __call__(
        self,
        X_hat: torch.Tensor,
        dimension_reduction_model: Optional[sklearn.decomposition.PCA] = None,
        final_timepoint_only: bool = True,
    ) -> pd.DataFrame:

        if X_hat.device != "cpu" and (final_timepoint_only):
            X_hat_ = self._DETACH_FINAL(X_hat)
        elif X_hat.device != "cpu":
            X_hat_ = self._DETACH(X_hat)
        elif final_timepoint_only:
            X_hat_ = X_hat[-1].numpy()
        else:
            X_hat_ = X_hat.numpy()

        if not dimension_reduction_model is None:
            X_hat_ = self._TRANSFORM(dimension_reduction_model, X_fin=X_hat_)

        return self._fate_df(self._query(X_fin=X_hat_))




import numpy as np
import torch


class AnnDataInspector:
    def __init__(self, adata):

        self._adata = adata

    @property
    def layers(self):
        return list(self._adata.layers)

    @property
    def obsm(self):
        return self._adata.obsm_keys()

    @property
    def attributes(self):
        return ["X"] + self.layers + self.obsm

    @property
    def is_np_array(self):
        return isinstance(self._X_use, np.ndarray)

    def as_tensor(self):
        return torch.Tensor(self._X_use)

    def _isolate_data_matrix(self, key):
        if key in self.layers:
            return self._adata.layers[key]
        elif key in self.obsm:
            return self._adata.obsm[key]

    def __call__(self, key, as_tensor=True):

        self._X_use = self._isolate_data_matrix(key)
        if self.is_np_array:
            if as_tensor:
                return self.as_tensor()
            return self._X_use



# -- import packages: ---------------------------------------------------------
import torch


# -- import local dependencies: -----------------------------------------------
from ....tools.utils import L2Norm


# -- MixIn class: -------------------------------------------------------------
class RegularizedVelocityRatioMixIn(object):
    """Mixin class for regularized velocity ratio operations.

    This class provides methods for computing regularized velocity ratio loss
    and incorporating it into the optimization step.

    Attributes:
        _L2Norm (L2Norm): Instance of L2Norm utility for computing L2 norm.

    Methods:
        __init__: Initialize the RegularizedVelocityRatioMixIn object.
        forward: Perform the forward step, integrating in latent space.
        velocity_ratio_loss_transform: Property to get the velocity ratio loss transformation method.
        _compute_magnitude_f: Compute the magnitude of drift.
        _compute_magnitude_g: Compute the magnitude of diffusion.
        _velocity_ratio_loss: Compute the velocity ratio loss.
        _dynamic_enforcement: Apply dynamic enforcement strategy.
        step: Perform a single optimization step incorporating velocity ratio loss.

    """

    def __init__(self):
        """Initialize the RegularizedVelocityRatioMixIn object."""
        super().__init__()
        
        self._L2Norm = L2Norm()


    def forward(self, Z0, t, **kwargs):
        """
        Perform the forward step: integrate in latent space.

        Args:
            Z0 (torch.Tensor): Initial latent space.
            t (torch.Tensor): Time steps.
            **kwargs: Additional keyword arguments.

        Returns:
            torch.Tensor: Resulting integrated latent space.

        """
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
    
    @property
    def velocity_ratio_loss_transform(self):
        """Property to get the velocity ratio loss transformation method.

        Returns:
            torch.function: Velocity ratio loss transformation method.

        """
        if not hasattr(self, "_velocity_ratio_loss_transform"):
            self._velocity_ratio_loss_transform = getattr(torch, self.hparams['velocity_ratio_params']['method'])
        return self._velocity_ratio_loss_transform
    
    def _compute_magnitude_f(self, Z_hat: torch.Tensor) -> torch.Tensor:
        """Compute the magnitude of drift.

        Args:
            Z_hat (torch.Tensor): Predicted latent space.

        Returns:
            torch.Tensor: Magnitude of drift.

        """
        return self._L2Norm(self.DiffEq.drift(Z_hat)).mean(1) * self.hparams['dt']
        
    def _compute_magnitude_g(self, Z_hat: torch.Tensor) -> torch.Tensor:
        """Compute the magnitude of diffusion.

        Args:
            Z_hat (torch.Tensor): Predicted latent space.

        Returns:
            torch.Tensor: Magnitude of diffusion.

        """
        return self._L2Norm(torch.stack([self.DiffEq.diffusion(Z) for Z in Z_hat]).squeeze(-1)).mean(1)

    def _velocity_ratio_loss(self, Z_hat: torch.Tensor) -> torch.Tensor:
        """Compute the velocity ratio loss.

        Args:
            Z_hat (torch.Tensor): Predicted latent space.

        Returns:
            torch.Tensor: Computed velocity ratio loss.

        """
        
        vr_params = self.hparams['velocity_ratio_params']
        
        self.reg_f = self._compute_magnitude_f(Z_hat)
        self.reg_g = self._compute_magnitude_g(Z_hat)
        
        return vr_params["enforce"] * self.velocity_ratio_loss_transform(
            vr_params["target"] - self.reg_f.div(self.reg_g)
        )
    
    def _dynamic_enforcement(self):

        """Apply dynamic enforcement strategy.
        
        From Anders: if you want to do the moving average, just grab the last N
        sinkhorn loss quantities from that list. and take the mean of that as your
        "velo_ratio_enforce." If its the first time computing the loss you can use
        velo_ratio_enforce = 0 instead of an average. this also means the VELO_RATIO_LOSS
        will change over training, decreasing as the model gets better at reconstructing
        the data.
        
        Returns:
            None

        """
        
        # TODO
        
        ...
        
    def step(self, batch, batch_idx, stage=None):
        
        """Perform a single optimization step incorporating velocity ratio loss.

        Args:
            batch: Input batch of data.
            batch_idx (int): Index of the batch.
            stage (str): Optional stage information.

        Returns:
            torch.Tensor: Total loss.
        """

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        self.sinkhorn_loss = self.compute_sinkhorn_divergence(
            X = batch.X, X_hat = X_hat, W = batch.W, W_hat = batch.W_hat
        )
        
        self.velocity_ratio_loss = self._velocity_ratio_loss(X_hat)
        
        self.total_loss = self.sinkhorn_loss + self.velocity_ratio_loss

        return self.total_loss.sum()




class BaseForwardMixIn(object):
    
    """
    Base class implementing forward mixing functionality.

    This mixin provides methods for forward mixing, which is a common operation
    in certain types of neural network models, particularly those related to
    sequence generation or transformation tasks.

    Attributes:
        None

    Methods:
        __init__: Initialize the BaseForwardMixIn object.
        forward: Perform the forward step, integrating in the latent space.
        step: Perform a single optimization step using the forward mixing approach.

    """
    
    def __init__(self) -> None:
        """
        Initialize the BaseForwardMixIn object.

        This initializes the object.

        Args:
            None

        Returns:
            None

        """
        super().__init__()


    def forward(self, Z0, t, **kwargs):
        """
        Perform the forward step: integrate in the latent space.

        This method integrates the given latent space Z0 over time t using
        specific parameters.

        Args:
            Z0 (tensor): Initial latent space.
            t (tensor): Time steps.
            **kwargs: Additional keyword arguments.

        Returns:
            tensor: Resulting integrated latent space.

        """
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
    
    def step(self, batch, batch_idx, stage=None):
        """Perform a single optimization step using the forward mixing approach.

        This method takes a batch of data and performs a single optimization step
        using the forward mixing approach. It calculates the Sinkhorn divergence loss
        between the actual and predicted outputs.

        Args:
            batch (Batch): Input batch of data.
            batch_idx (int): Index of the batch.
            stage (str): Optional stage information.

        Returns:
            tensor: Computed Sinkhorn divergence loss.

        """
        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        self.sinkhorn_loss = self.compute_sinkhorn_divergence(
            X = batch.X, X_hat = X_hat, W = batch.W, W_hat = batch.W_hat
        )
        return self.sinkhorn_loss.sum()



# -- import packages: --------------------------------------------------------
import ABCParse
import anndata
import logging

# -- import local dependencies: -----------------------------------------------
from ... import tools

# -- set type hints: ----------------------------------------------------------
from typing import Optional


# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- mix-in cls: --------------------------------------------------------------
class kNNMixIn(ABCParse.ABCParse, object):
    def __init__(self, *args, **kwargs) -> None:
        self.__parse__(locals())

    @property
    def _adata_kNN_fit(self):
        return self.adata[self.adata.obs[self._kNN_fit_subset]].copy()

    def _kNN_info_msg(self) -> None:

        logger.info(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_fit_subset}']")

    def configure_kNN(
        self,
        adata: Optional[anndata.AnnData] = None,
        kNN_key: Optional[str] = None,
        kNN_fit_subset: Optional[str] = None,
    ) -> None:
        """
        subset key should point to a col in adata.obs of bool vals
        """
        self.__update__(locals())

        logger.info(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_fit_subset}']")

        self._kNN = tools.kNN(
            adata=self._adata_kNN_fit,
            use_key=self._kNN_key,
        )

    @property
    def kNN(self) -> tools.kNN:
        if not hasattr(self, "_kNN"):
            self.configure_kNN(
                adata=self._adata_kNN_fit,
                kNN_key=self._kNN_key,
                kNN_fit_subset=self._kNN_fit_subset,
            )
        return self._kNN




"""
Flexible fetch of optimizer, LR_scheduler, other functions.

"""

# -- import packages: ----------------------------------------------------------
import torch

# -- import local dependencies: ------------------------------------------------
import ABCParse


# -- define types: -------------------------------------------------------------
from typing import Union, Any
NoneType = type(None)


class FunctionFetch(ABCParse.ABCParse):
    """Fetch a function from a flexible input."""
    def __init__(self, module=None, parent=None):
        self.__parse__(locals())

    def __call__(self, func: Union[Any, str]):
        if isinstance(func, str):
            return getattr(self.module, func)
        elif issubclass(func, self.parent):
            return func
        else:
            print("pass something else....")
            
# -- supporting functions: -----------------------------------------------------
def fetch_optimizer(func):
    """
    Examples:
    ---------
    fetch_optimizer("RMSprop")
    >>> torch.optim.rmsprop.RMSprop

    fetch_optimizer(torch.optim.RMSprop)
    >>> torch.optim.rmsprop.RMSprop
    """
    fetch = FunctionFetch(module=torch.optim, parent=torch.optim.Optimizer)
    return fetch(func)


def fetch_lr_scheduler(func):
    """
    Examples:
    ---------
    fetch_lr_scheduler("StepLR")
    >>> torch.optim.lr_scheduler.StepLR

    fetch_lr_scheduler(torch.optim.lr_scheduler.StepLR)
    >>> torch.optim.lr_scheduler.StepLR
    """
    fetch = FunctionFetch(
        module=torch.optim.lr_scheduler, parent=torch.optim.lr_scheduler.LRScheduler
    )
    return fetch(func)


from typing import Union
import numpy, torch


def sum_normalize(X: Union[torch.Tensor, numpy.ndarray], sample_axis=1):
    """
    Parameters:
    -----------
    X
        torch.Tensor or nump.array

    sample_axis
        default: 1
        type: int

    Returns:
    --------
    return X_
    """
    return X / X.sum(sample_axis)[:, None]


from ._intermittent_saves import IntermittentSaves
from ._gradient_potential_callback import GradientPotentialTest
from ._testing import Testing
from ._model_logging import ModelLogging
from ._memory_monitor import MemoryMonitor
from ._basic_progress_bar import BasicProgressBar




# -- import packages: ----------------------------------------------------------
import numpy as np
import anndata
import ABCParse


# -- import local dependencies: ------------------------------------------------
from ...core import utils


class MeshGrid(ABCParse.ABCParse):
    def __init__(self, adata: anndata.AnnData, gridpoints: int = 250, use_key: str = "X_umap"):

        self.__parse__(locals(), public=[None])

    @property
    def X_umap(self):
        return self._adata.obsm[self._use_key]

    @property
    def _X_MIN(self):
        return np.min(self.X_umap[:, 0])

    @property
    def _X_MAX(self):
        return np.max(self.X_umap[:, 0])

    @property
    def _Y_MIN(self):
        return np.min(self.X_umap[:, 1])

    @property
    def _Y_MAX(self):
        return np.max(self.X_umap[:, 0])

    @property
    def _X_GRID(self):
        return np.linspace(self._X_MIN, self._X_MAX, self._gridpoints)

    @property
    def _Y_GRID(self):
        return np.linspace(self._Y_MIN, self._Y_MAX, self._gridpoints)

    def __call__(self):
        return np.meshgrid(self._X_GRID, self._Y_GRID)


# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib


# -- import local dependencies: -----------------------------------------------
from ._version import Version


# -- set typing: --------------------------------------------------------------
from typing import Dict, Union


# -- operational class: -------------------------------------------------------
class Project(ABCParse.ABCParse):
    """Object container for an scDiffEq project

    Attributes:
        path (Union[str, pathlib.Path]): Path to the project, created by scDiffeq.
        metrics_groupby (str): Grouping method for metrics, default is "epoch".
    """

    def __init__(
        self,
        path: Union[str, pathlib.Path] = pathlib.Path("./").absolute(),
        metrics_groupby: str = "epoch",
        *args,
        **kwargs,
    ) -> None:
        """Initialize the project object by providing a path.

        Args:
            path (Union[str, pathlib.Path], optional): Path to the project, created by scDiffeq, by default pathlib.Path("./").absolute()
            metrics_groupby (str, optional): Grouping method for metrics, by default "epoch"

        Returns:
            None
        """
        self.__parse__(locals())

        self._set_version_path_attributes()

    @property
    def _PROJECT_PATH(self) -> pathlib.Path:
        """Check and format the provided project path

        Returns:
            pathlib.Path
            Formatted project path

        Raises:
            TypeError: If the path is not of type pathlib.Path or str
        """
        if isinstance(self._path, pathlib.Path):
            return self._path
        elif isinstance(self._path, str):
            return pathlib.Path(self._path)
        else:
            raise TypeError("arg: `path` must be of type: [pathlib.Path, str]")

    @property
    def _VERSION_PATHS(self) -> Dict[str, pathlib.Path]:
        """Assemble available version paths based on the provided project path

        Returns:
            Dict[str, pathlib.Path]
            Dictionary of version names and their corresponding paths
        """
        version_paths = sorted(list(self._PROJECT_PATH.glob("version_*")))
        return {path.name: path for path in version_paths}

    def _set_version_path_attributes(self) -> None:
        """Sets the name of each version as a class attribute, pointing to the
        path of the version.
        """
        for v_name, v_path in self._VERSION_PATHS.items():
            version = Version(path=v_path, groupby=self._metrics_groupby)
            setattr(self, v_name, version)

    def __getitem__(self, version: int) -> pathlib.Path:
        """Format version key and return version

        Args:
            version (int): Version number

        Returns:
            pathlib.Path
            Path to the specified version
        """
        return getattr(self, f"version_{version}")

    def __repr__(self) -> str:
        return """scDiffEq Project"""




from ..core import utils

import adata_query
import ABCParse
import anndata
import pandas as pd
import numpy as np
import annoy


from typing import List, Optional


class kNN(ABCParse.ABCParse):
    """kNN container"""
    _IDX_BUILT = False

    def __init__(
        self,
        adata: anndata.AnnData,
        use_key: str = "X_pca",
        metric: str = "euclidean",
        n_neighbors: int = 20,
        n_trees: int = 10,
        *args,
        **kwargs,
    ):
        """Initialize kNN
        
        Extended description.
        
        Args:
            adata (anndata): Description.
            
            use_key (str): Description. **Default**: ``"X_pca"``.
        
        Returns:
            None
        """
        self.__parse__(locals(), public=[None])
        self.idx = self._build()

    @property
    def X_use(self):
        if not hasattr(self, "_X_use"):
            self._X_use = adata_query.fetch(self._adata, key=self._use_key, torch=False)
        return self._X_use

    @property
    def _N_CELLS(self):
        return self.X_use.shape[0]

    @property
    def _N_DIMS(self):
        return self.X_use.shape[1]

    def _build(self) -> annoy.AnnoyIndex:
        idx = annoy.AnnoyIndex(self._N_DIMS, self._metric)
        [idx.add_item(i, self.X_use[i]) for i in range(self._N_CELLS)]
        idx.build(self._n_trees)
        return idx

    def query(self, X_query: np.ndarray) -> np.ndarray:
        return np.array(
            [
                self.idx.get_nns_by_vector(X_query[i], self._n_neighbors)
                for i in range(X_query.shape[0])
            ]
        )

    def _count_values(self, col: pd.Series) -> dict:
        return col.value_counts().to_dict()

    def _max_count(self, col: pd.Series) -> str:
        return col.value_counts().idxmax()

    def count(
        self,
        query_result,
        obs_key: str,
        max_only: bool = False,
        n_neighbors: Optional[int] = None,
    ):
        if n_neighbors is None:
            n_neighbors = self._n_neighbors

        nn_adata = self._adata[query_result.flatten()]
        query_df = pd.DataFrame(
            nn_adata.obs[obs_key].to_numpy().reshape(-1, n_neighbors).T
        )
        del nn_adata

        if not max_only:
            return [
                self._count_values(query_df[i]) for i in query_df.columns
            ]  # list of dicts
        return [
            self._max_count(query_df[i]) for i in query_df.columns
        ]  # list of values

    def aggregate(
        self,
        X_query,
        obs_key: str,
        max_only: bool = False,
        n_neighbors: Optional[int] = None,
    ):
        _df = (
            pd.DataFrame(
                self.count(
                    query_result=self.query(X_query=X_query),
                    obs_key=obs_key,
                    max_only=max_only,
                    n_neighbors=n_neighbors,
                )
            )
            .fillna(0)
            .sort_index(axis=1)
        )
        if not max_only:
            return _df
        return _df.rename({0: obs_key}, axis=1)

    def multi_aggregate(
        self,
        X_query,
        obs_key: str,
        max_only: bool = False,
        n_neighbors: Optional[int] = None,
    ) -> List[pd.DataFrame]:
        _list_of_dfs = [
            self.aggregate(
                X_query=X_query[i],
                obs_key=obs_key,
                max_only=max_only,
                n_neighbors=n_neighbors,
            )
            for i in range(len(X_query))
        ]

        if max_only:
            concat_df = pd.concat(_list_of_dfs, axis=1)
            concat_df.columns = range(len(X_query))
            return concat_df

        return _list_of_dfs

    def __repr__(self):
        return f"kNN: {self._use_key}"



# -- import: ------------------------------------------------------------------
from ._function_fetch import FunctionFetch, fetch_optimizer, fetch_lr_scheduler
from ._function_kwargs import extract_func_kwargs, function_kwargs
from ._logging_learnable_hparams import LoggingLearnableHParams
from ._sum_normalize import sum_normalize
from ._logger_bridge import LoggerBridge
from ._normalize_time_to_range import normalize_time
from ._anndata_inspector import AnnDataInspector
from ._fast_graph import FastGraph
from ._idx_to_int_str import idx_to_int_str
from ._logs import Logs, PretrainLogs, TrainLogs
from ._filter_df import filter_df
from ._knn_graph_query import kNNGraphQuery
from ._flexible_component_loader import FlexibleComponentLoader

# -- export: ------------------------------------------------------------------
__all__ = [
    "FunctionFetch",
    "fetch_optimizer",
    "fetch_lr_scheduler",
    "extract_func_kwargs",
    "function_kwargs",
    "LoggingLearnableHParams",
    "sum_normalize",
    "LoggerBridge",
    "normalize_time",
    "AnnDataInspector",
    "FastGraph",
    "idx_to_int_str",
    "Logs",
    "PretrainLogs",
    "TrainLogs",
    "filter_df",
    "kNNGraphQuery",
    "FlexibleComponentLoader",
]



# -- import packages: ---------------------------------------------------------
import autodevice
import logging
import pathlib
import torch

# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- mix-in cls: --------------------------------------------------------------
class UtilityIOMixIn(object):
    """MixIn container for various model I/O functions."""

    def to(self, device: torch.device) -> None:
        """Send model to a specific device"""
        self.DiffEq = self.DiffEq.to(device)

    def freeze(self) -> None:
        """Freeze lightning model"""
        self.DiffEq.freeze()

    def load(
        self,
        ckpt_path: Union[str, pathlib.Path],
        freeze: Optional[bool] = True,
        device: Optional[torch.device] = autodevice.AutoDevice(),
    ) -> None:
        """Load a model from a path to a checkpoint."""

        self.__update__(locals())

        self.DiffEq = self.DiffEq.load_from_checkpoint(self._ckpt_path)
        self.DiffEq = self.DiffEq.to(self._device)

        if freeze:
            self.freeze()




from lightning import Callback
import pickle
import os


class IntermittentSaves(Callback):
    def __init__(self, frequency=50):
                
        self.frequency = frequency

    def on_train_batch_end(self, trainer, model, outputs, batch, batch_idx, *args, **kwargs):
        
        """ """
                
        epoch = model.current_epoch
        
        if (epoch % self.frequency == 0) and (batch_idx == 0):
            dpath = os.path.join(trainer.logger.log_dir, "ckpt_outputs")
            if not os.path.exists(dpath):
                os.mkdir(dpath)

            fpath = os.path.join(dpath, "epoch_{}.batch_{}.pkl".format(epoch, batch_idx))
            pickle.dump(obj=outputs,
                        file=open(fpath, "wb"),
                        protocol=pickle.HIGHEST_PROTOCOL,
                       )



# -- import packages: ---------------------------------------------------------
import neural_diffeqs
import torch

# -- import local dependencies: -----------------------------------------------
from . import base, mix_ins

# -- set type hints: ----------------------------------------------------------
from typing import Literal, Optional, Union, List

# -- lightning model: ---------------------------------------------------------
class LightningSDE_VAE_FateBiasAware(
    mix_ins.VAEMixIn,
    mix_ins.FateBiasVAEMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    """LightningSDE-VAE-FateBiasAware"""
    def __init__(
        self,
        data_dim: int,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr: float = 1e-5,
        pretrain_lr: float = 1e-3,
        pretrain_epochs: int = 100,
        pretrain_optimizer: torch.optim.Optimizer = torch.optim.Adam,
        train_optimizer: torch.optim.Optimizer = torch.optim.RMSprop,
        pretrain_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        train_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.StepLR,
        pretrain_step_size: int = 200,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint: bool = False,
        backend: str = "auto",
        # -- sde params: -------------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        sigma_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type="ito",
        noise_type="general",
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        # -- encoder parameters: -----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- decoder parameters: -----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        # -- fate bias parameters: ---------------------------------------------
        t0_idx=None,
        kNN_Graph=None,
        fate_bias_csv_path: Optional[str]=None,
        fate_bias_multiplier: float = 1,
        PCA=None,
        loading_existing: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        LightningSDE_VAE_FateBiasAware

        Parameters
        ----------
        latent_dim : int, optional
            Dimensionality of the latent space, by default 50.
        name : str, optional
            Name of the model, by default None.
        mu_hidden : Union[List[int], int], optional
            Hidden layer sizes for the drift neural network, by default [400, 400, 400].
        sigma_hidden : Union[List[int], int], optional
            Hidden layer sizes for the diffusion neural network, by default [400, 400, 400].
        mu_activation : Union[str, List[str]], optional
            Activation function(s) for the drift neural network, by default 'LeakyReLU'.
        sigma_activation : Union[str, List[str]], optional
            Activation function(s) for the diffusion neural network, by default 'LeakyReLU'.
        mu_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the drift neural network, by default 0.1.
        sigma_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the diffusion neural network, by default 0.1.
        mu_bias : bool, optional
            Whether to use bias in the drift neural network, by default True.
        sigma_bias : List[bool], optional
            Whether to use bias in the diffusion neural network, by default True.
        mu_output_bias : bool, optional
            Whether to use bias in the output layer of the drift neural network, by default True.
        sigma_output_bias : bool, optional
            Whether to use bias in the output layer of the diffusion neural network, by default True.
        mu_n_augment : int, optional
            Number of augmentations for the drift neural network, by default 0.
        sigma_n_augment : int, optional
            Number of augmentations for the diffusion neural network, by default 0.
        sde_type : str, optional
            Type of stochastic differential equation, by default 'ito'.
        noise_type : str, optional
            Type of noise, by default 'general'.
        brownian_dim : int, optional
            Dimensionality of the Brownian motion, by default 1.
        coef_drift : float, optional
            Coefficient of drift, by default 1.0.
        coef_diffusion : float, optional
            Coefficient of diffusion, by default 1.0.
        encoder_n_hidden : int, optional
            Number of hidden layers in the encoder, by default 4.
        encoder_power : float, optional
            Power of the encoder, by default 2.
        encoder_activation : Union[str, List[str]], optional
            Activation function(s) for the encoder, by default 'LeakyReLU'.
        encoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the encoder, by default 0.2.
        encoder_bias : bool, optional
            Whether to use bias in the encoder, by default True.
        encoder_output_bias : bool, optional
            Whether to use bias in the output layer of the encoder, by default True.
        decoder_n_hidden : int, optional
            Number of hidden layers in the decoder, by default 4.
        decoder_power : float, optional
            Power of the decoder, by default 2.
        decoder_activation : Union[str, List[str]], optional
            Activation function(s) for the decoder, by default 'LeakyReLU'.
        decoder_dropout : Union[float, List[float]], optional
            Dropout rate(s) for the decoder, by default 0.2.
        decoder_bias : bool, optional
            Whether to use bias in the decoder, by default True.
        decoder_output_bias : bool, optional
            Whether to use bias in the output layer of the decoder, by default True.
        t0_idx : int, optional
            Initial time index for fate bias, by default None.
        kNN_Graph : object, optional
            k-Nearest Neighbors graph for fate bias, by default None.
        fate_bias_csv_path : str, optional
            Path to the CSV file containing fate bias information, by default None.
        fate_bias_multiplier : float, optional
            Multiplier for fate bias, by default 1.
        PCA : object, optional
            PCA object for dimensionality reduction, by default None.
        loading_existing : bool, optional
            Whether to load an existing model, by default False.

        Returns
        -------
        None

        Notes
        -----
        This class implements a VAE with fate bias aware SDE using PyTorch Lightning.

        Examples
        --------
        >>> model = LightningSDE_VAE_FateBiasAware(latent_dim=50, dt=0.1)
        >>> model.fit(data)
        """
        super().__init__()

        name = self._configure_name(name, loading_existing=loading_existing)

        self.save_hyperparameters(ignore=["kNN_Graph"])

        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=neural_diffeqs.NeuralSDE, kwargs=locals())
        self._configure_lightning_model(kwargs=locals())

        self._configure_fate(
            graph=kNN_Graph,
            csv_path=fate_bias_csv_path,
            t0_idx=t0_idx,
            fate_bias_multiplier=fate_bias_multiplier,
            PCA=PCA,
        )

    def __repr__(self) -> Literal['LightningSDE-VAE-FateBiasAware']:
        return "LightningSDE-VAE-FateBiasAware"



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import logging
import os
import numpy as np
import pathlib


# -- import local dependencies: -----------------------------------------------
from ._figshare_downloader import figshare_downloader

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)


# -- preprocessing dataset cls: -----------------------------------------------
class UnProcessedHumanHematpoiesisDataHandler(ABCParse.ABCParse):
    # HTTPS = "https://www.dropbox.com/s/8m8n6fj8yn1ryjd/hsc_all_combined_all_layers.h5ad?dl=1"
    RAW_FNAME = "_hsc_all_combined_all_layers.h5ad"
    PROCESSED_FNAME = "human_hematopoiesis.processed.h5ad"

    def __init__(
        self,
        data_dir=os.getcwd(),
        skip_scaling: bool = False,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("human_hematopoiesis")
        if not path.exists():
            path.mkdir()
        return path

    @property
    def raw_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.RAW_FNAME)

    @property
    def processed_h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self.PROCESSED_FNAME)

    def download_raw(self):
        if not self.raw_h5ad_path.exists():
            figshare_downloader(
                figshare_id=54154235,
                write_path=self.raw_h5ad_path,
            )

    @property
    def raw_adata(self):
        if not hasattr(self, "_raw_adata"):
            if not self.raw_h5ad_path.exists():
                self.download_raw()
            self._raw_adata = anndata.read_h5ad(self.raw_h5ad_path)
            self._raw_idx = self._raw_adata.obs.index
        return self._raw_adata

    @property
    def pp_adata(self):
        if not hasattr(self, "_pp_adata"):
            if not self.processed_h5ad_path.exists():
                self._pp_adata = self.get(skip_scaling=self._skip_scaling)
            else:
                self._pp_adata = anndata.read_h5ad(self.processed_h5ad_path)
        return self._pp_adata

    def seurat_preprocessing(self, adata: anndata.AnnData):

        import scanpy as sc

        adata = adata.copy()

        adata.obs["time"] = 3
        adata.obs.loc[adata.obs.batch == "JL_10", "time"] = 5
        adata.obs.time.value_counts()
        adata.obs_names_make_unique()
        adata.obs["nGenes"] = (adata.X > 0).sum(1)
        adata.obs["UMI"] = (adata.X).sum(1)
        adata.layers["X_orig"] = adata.X
        adata.X = adata.layers["labeled_TC"].copy()
        # Convert pandas Series to numpy array for proper boolean indexing
        jl_10_mask = (adata.obs.batch == "JL_10").values
        adata.X[jl_10_mask, :] *= 3 / 5
        adata.obs.rename({"time": "_time"}, axis=1, inplace=True)
        t_map = {"JL_10": 4, "JL12_0": 7, "JL12_1": 7}
        adata.obs["t"] = adata.obs["batch"].map(t_map)
        sc.pp.recipe_seurat(adata)

        return adata

    def _match_and_filter_on_idx(self, adata: anndata.AnnData):
        """Match and filter on idx."""

        # HTTPS = "https://www.dropbox.com/s/n9mx9trv1h78q0r/hematopoiesis_v1.h5ad?dl=1"
        _secondary_h5ad = self.data_dir.joinpath("_dynamo_hematopoiesis_v1.h5ad")
        if not _secondary_h5ad.exists():
            figshare_downloader(
                figshare_id=54154238,
                write_path=_secondary_h5ad,
            )

        adata_dyn = anndata.read_h5ad(_secondary_h5ad)

        match_idx = np.where(adata.obs.index.isin(adata_dyn.obs.index))
        adata = adata[match_idx].copy()
        adata.obs["cell_type"] = adata_dyn[adata.obs.index].obs["cell_type"].values

        return adata

    def dimension_reduction(self, adata: anndata.AnnData, skip_scaling: bool = True):

        import sklearn
        import umap

        from .. import io

        adata = adata.copy()

        self.SCALER_MODEL = sklearn.preprocessing.StandardScaler()
        self.PCA_MODEL = sklearn.decomposition.PCA(n_components=50)
        self.UMAP_MODEL = umap.UMAP(
            n_components=2,
            n_neighbors=25,
            random_state=1,
            min_dist=0.5,
        )

        if skip_scaling:
            adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.X)
        else:
            adata.layers["X_scaled"] = self.SCALER_MODEL.fit_transform(adata.X)
            adata.obsm["X_pca"] = self.PCA_MODEL.fit_transform(adata.layers["X_scaled"])
            io.write_pickle(
                self.SCALER_MODEL,
                self.data_dir.joinpath("human_hematopoiesis.scaler.pkl"),
            )

        adata.obsm["X_umap"] = self.UMAP_MODEL.fit_transform(
            adata.obsm["X_pca"][:, :10]
        )

        io.write_pickle(
            self.PCA_MODEL, self.data_dir.joinpath("human_hematopoiesis.pca.pkl")
        )
        io.write_pickle(
            self.UMAP_MODEL,
            self.data_dir.joinpath("human_hematopoiesis.umap.pkl"),
        )

        return self._match_and_filter_on_idx(adata)

    def get(self, skip_scaling: bool = False):

        adata_pp = self.seurat_preprocessing(self.raw_adata)
        adata_pp = self.dimension_reduction(adata_pp, skip_scaling=skip_scaling)
        adata_pp.write_h5ad(self.processed_h5ad_path)
        return adata_pp


# -- dataset cls: -------------------------------------------------------------
class HumanHematopoiesisDataset(ABCParse.ABCParse):
    _FNAME = "human_hematopoiesis"
    _ADATA_FNAME = f"{_FNAME}.processed.h5ad"
    figshare_ids = {
        "adata": {"file_id": 54154232, "fname": _ADATA_FNAME},
        "scaler": {"file_id": 54154226, "fname": f"{_FNAME}.scaler.pkl"},
        "pca": {"file_id": 54154223, "fname": f"{_FNAME}.pca.pkl"},
        "umap": {"file_id": 54154229, "fname": f"{_FNAME}.umap.pkl"},
    }

    def __init__(
        self,
        data_dir=os.getcwd(),
        force_download: bool = False,
        *args,
        **kwargs,
    ):

        self.__parse__(locals())

    @property
    def _scdiffeq_parent_data_dir(self):
        path = pathlib.Path(self._data_dir).joinpath("scdiffeq_data")
        path.mkdir(exist_ok=True, parents=True)
        return path

    @property
    def data_dir(self):
        path = self._scdiffeq_parent_data_dir.joinpath("human_hematopoiesis")
        path.mkdir(exist_ok=True, parents=True)
        return path

    @property
    def h5ad_path(self) -> pathlib.Path:
        return self.data_dir.joinpath(self._ADATA_FNAME)

    def download(self):
        for _, val in self.figshare_ids.items():
            figshare_downloader(
                figshare_id=val["file_id"],
                write_path=self.data_dir.joinpath(val["fname"]),
            )

    @property
    def adata(self) -> anndata.AnnData:
        if not hasattr(self, "_adata"):
            if (not self.h5ad_path.exists()) or self._force_download:
                self.download()
            self._adata = anndata.read_h5ad(self.h5ad_path)
        return self._adata


# -- download function for unprocessed data: -----------------------------------
def _download_unprocessed_human_hematopoiesis(
    data_dir=os.getcwd(),
    skip_scaling: bool = False,
):
    _handler = UnProcessedHumanHematpoiesisDataHandler(
        data_dir=data_dir,
        skip_scaling=skip_scaling,
    )
    return _handler.pp_adata


# -- main api-facing function: ------------------------------------------------
def human_hematopoiesis(
    data_dir: str = os.getcwd(),
    skip_scaling: bool = False,
    force_download: bool = False,
    download_unprocessed: bool = False,
) -> anndata.AnnData:
    """Human hematopoiesis dataset

    Args:
        data_dir: str, default=os.getcwd()
            Path to the directory where the data will be saved.
        skip_scaling: bool, default=False
            Whether to skip scaling.
        force_download: bool, default=False
            Whether to force download the data.
        download_unprocessed: bool, default=False
            Whether to download the unprocessed data.

    Returns:
        anndata.AnnData: Preprocessed AnnData object.
    """
    if download_unprocessed:
        return _download_unprocessed_human_hematopoiesis(
            data_dir=data_dir,
            skip_scaling=skip_scaling,
        )
    else:
        data_handler = HumanHematopoiesisDataset(
            data_dir=data_dir,
            force_download=force_download,
        )
        return data_handler.adata
